//
// parseHook.cpp
//
// Copyright Howard Berkey, 1997
//
// Take the passed-in string, parse it for selected non-shell type stuff,
// replace it inline if necessary with the correct shell commands, or 
// handle it and skip the popen step altogether.
//
// As far as a parser goes, this code is intended to be pretty simplistic.
// A more "correct" way to do this is in C++, instantiating the correct parser 
// object using something like a parser factory object that looks at the parse
// tag (the first few characters of the input) and returns a reference to the
// correct parser object.  However, for simplicity I'm sticking with pretty much 
// vanilla C code here, since this is a BeIDE add-on example, not a parser example 
// :-)
//


#include "Executioner.h"
#include <AppKit.h>
#include <StorageKit.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <mmemory.h>    // auto_ptr


//
// parser function type decl.  return true if we didn't 
// recognize the text in the buffer (and the search should 
// continue), false if we take care of the buffer in the hook. 
// If all parsefuncs returned true, 
//
typedef bool (*executioner_parsefunc)(MTextAddOn *, char**);

//
// parse table element decl
//
struct executioner_parsetags {
	char *tag;
	executioner_parsefunc func;
};

//
// Make our parse table.  Currently there is only one parser.
// (well, two, if you count the version easter-egg)
//
// mwcc compiler errors currently look like this:
//
//### /boot/apps/Metrowerks/tools/mwcc Compiler Error:
//#	#include "foo.h" 
//#	                     ^
//# the file 'foo.h' cannot be opened
//#----------------------------------------------------------
//File "/boot/home/howard/foobar/foo.cpp"; Line 3
//#----------------------------------------------------------
//
// If we detect either the ### tag or the File " tag we attempt to
// parse the buffer as a mwcc error buffer.
//
// Other parse tags can be added here in the future.
//

static bool mwcc_error(MTextAddOn *mtex, char **cmdbuf);
static bool print_version(MTextAddOn *mtex, char **cmdbuf);
static bool posixtool_check(MTextAddOn *mtex, char **cmdbuf);

static struct executioner_parsetags p_tags[] = {
	{"Executioner: version", print_version},
	{"File ", mwcc_error}, 
	{"###", mwcc_error}, 
	{"", posixtool_check},    // is it "filename:69:"?
	{ NULL, NULL}
};



//
// Parsing hook function, called from the perform_edit exported function
//
bool parseHook(MTextAddOn *mtex, char **cmdbuf)
{
	int i;
	bool rc = true;

	for(i=0; (p_tags[i]).tag != NULL; i++)
	{
		if(strncmp(*cmdbuf, (p_tags[i]).tag, strlen((p_tags[i]).tag)) == 0)
			if((p_tags[i]).func != NULL)
				if (false == (rc = (p_tags[i]).func(mtex, cmdbuf)))
					break;
	}

	return rc;
}


//
// specific parser hooks go here.  
//

//
// prints version string
//
bool print_version(MTextAddOn *mtex, char ** /*cmdbuf*/)
{
	long s_start, s_end;
	
	mtex->GetSelection(&s_start, &s_end);		
	mtex->Select(s_end, s_end);
	mtex->Insert(EXECUTIONER_VERSION);	
	s_start = s_end;
	s_end = s_start + strlen(EXECUTIONER_VERSION);
	mtex->Select(s_start, s_end);
	return false;
}


//****************************************************************************
// mwcc error parser
//****************************************************************************

//
// mwtoken is a little embarrassing.  I reverse-engineered the BeIDE Token type
// used in selecting a source line.  In my defense, Jon Watte suggested it :-)
// The obligatory warning:  This is not supported by MetroWerks and will someday
// not work under future (post-1.3.2) BeIDE's.  Hopefully by then MetroWerks 
// will either release more comprehensive docs on BMessaging the BeIDE, or
// they will provide "Go To Line" functionality in the plugin API for arbitrary 
// BeIDE windows.  Until then, this works just fine.
// 


static char filebuf[4096];     // so we don't leak memory anymore

struct mwtoken {
	uint32 line;
	uint32 position;
	uint32 pad;
	uint32 pad1f;
	uint32 pad10;
	char text[32];
	uint32 zero;
};

bool mwcc_error(MTextAddOn *mtex, char **cmdbuf)
{
	char *tmp;
	int32 line;
	int i;
	entry_ref ref;
	status_t rc;
	
	
	//
	// skip to the first file in our debug output if not there already
	//
	if((tmp = strstr(*cmdbuf, "File \"")) == NULL)
	{
		return true;
	}

	//
	// Skip to the open quote and go one more position
	//
	if((tmp = strchr(tmp, '\"')) == NULL)
	{
		return true;
	}
	
	if(*tmp == '\"')
		tmp++;
	
	if(tmp == '\0' || tmp >= *cmdbuf + strlen(*cmdbuf))
	{
		return true;
	}

	//
	// copy the file path name into our own buffer up to the closing quote
	//	
	i=0;

	while(*tmp != '\"' && *tmp != '\0')
	{
		filebuf[i] = *tmp;
		i++;
		tmp++;
	}
	
	filebuf[i] = '\0';

	//
	// Get the entry_ref for the file the error occurred in
	//	
	BEntry entry(filebuf);
	
	if(entry.GetRef(&ref) != B_NO_ERROR)
	{
		return true;	
	}

	//
	// Get the line number
	//
	if(*tmp == '\0' || (tmp = strstr(tmp, "Line ")) == NULL)
	{
		return true;
	}
	
	// if here tmp points at "Line "
	
	tmp += 5; // strlen("Line ")
	
	line = atol(tmp);


	//
	// Make our REFS_RECEIVED message
	//	
	BMessage msg(B_REFS_RECEIVED);

	if(msg.AddRef("refs", &ref) != B_OK)
		return true;	

	//
	// Fill out our hacked-up BeIDE Token type that is added to the 
	// REFS_RECEIVED message by the BeIDE when you dbl-click in the Message
	// window
	//
	struct mwtoken hack;
	memset(&hack, 0, sizeof(hack));
	hack.line = line - 1;
	hack.pad1f = 0x0000001f;
	hack.pad10 = 0x00000010;
	if(msg.AddData("Token", 'Toke', &hack, 56) != B_OK)
		return true;	

	//
	// Send the message
	//	
	mtex->Delete();
	be_app->Lock();
	rc = be_app->PostMessage(&msg);
	be_app->Unlock();

	return false;
}

//****************************************************************************
// posix tool output check 
//
// This works with NO prefix text (it just tests). If the text is 
// "XXX:###:zzz", and ### is all digits, then we assume we got it correctly, 
// and that XXX is the filename, and ### is the linenumber. So this will break
// if there is a ':' in the filename.
//****************************************************************************

bool posixtool_check(MTextAddOn *mtex, char **cmdbuf)
{
	char * tmp, * tmp2;
	int32 line, filenamelen=0, linenumstrlen=0;
	int i;
	entry_ref ref;
	status_t rc;
	
	
	if((tmp = strchr(*cmdbuf, ':')) == NULL)
		return true;
	
	// is a newline before the ':'?
	tmp2 = strchr(*cmdbuf, '\n');
	if (tmp2 && tmp>tmp2)
		return true;
	
	// copy up to the ':'
	filenamelen = tmp - *cmdbuf;		
	strncpy(filebuf, *cmdbuf, filenamelen);
	filebuf[filenamelen] = '\0';
	//printf("%d \"%s\"\n",filenamelen,filebuf);


	//
	// Get the line number
	//

	// find second ':'
	if((tmp = strchr((*cmdbuf)+filenamelen+2, ':')) == NULL)
		return true;

	linenumstrlen = (tmp - *cmdbuf) - filenamelen - 1;
	tmp -= linenumstrlen;
	
	for (i=0; i<linenumstrlen; i++)
		if (!isdigit(tmp[i]))
			return true;	

	line = atol(tmp);
	//printf("%d\n", line);



	//
	// Get the entry_ref for the file
	//	
	BEntry entry(filebuf);
	
	if(entry.GetRef(&ref) != B_NO_ERROR)
		return true;	


	//
	// Make our REFS_RECEIVED message
	//	
	BMessage msg(B_REFS_RECEIVED);

	if(msg.AddRef("refs", &ref) != B_OK)
		return true;	

	//
	// Fill out our hacked-up BeIDE Token type that is added to the 
	// REFS_RECEIVED message by the BeIDE when you dbl-click in the Message
	// window
	//
	struct mwtoken hack;
	memset(&hack, 0, sizeof(hack));
	hack.line = line - 1;
	hack.pad1f = 0x0000001f;
	hack.pad10 = 0x00000010;
	if(msg.AddData("Token", 'Toke', &hack, 56) != B_OK)
		return true;	



	//
	// Adjust the selection to no longer include the line we just parsed
	//

	const char *txtbuf;
	txtbuf = mtex->Text();

	long s_start, s_end;
	mtex->GetSelection(&s_start, &s_end);
	while(s_start < mtex->TextLength() && txtbuf[s_start] != '\n')
		s_start++;

	if (txtbuf[s_start] == '\n')
		s_start++;

	mtex->Select(s_start, s_end);


	//
	// Send the message
	//	
	be_app->Lock();
	rc = be_app->PostMessage(&msg);
	be_app->Unlock();

	return false;
}


